/* Webcamoid, webcam capture application.
 * Copyright (C) 2023  Gonzalo Exequiel Pedone
 *
 * Webcamoid is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Webcamoid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Webcamoid. If not, see <http://www.gnu.org/licenses/>.
 *
 * Web-Site: http://webcamoid.github.io/
 */

package org.webcamoid.plugins.AudioDevice.submodules.androidaudio;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;

public class AkAndroidAudioCallbacks extends BroadcastReceiver
{
    private long m_userPtr = 0;

    private AkAndroidAudioCallbacks(long userPtr)
    {
        m_userPtr = userPtr;
    }

    // BroadcastReceiver

    @Override
    public void onReceive(Context context, Intent intent)
    {
        devicesUpdated(m_userPtr);
    }

    private static native void devicesUpdated(long userPtr);
}
